use serde::Serialize;
use serde_json::json;
use rocket::http::Status;
use rocket::request::{Request, Outcome};
use rocket::response::{self, Response, Responder};
use rocket::serde::json::Json;
use serde_json::Value;
use diesel::result::Error as DieselError;
use argon2::password_hash::errors::Error as PasswordHashError;

use crate::dns::ConnectorError;
use crate::models;

#[derive(Debug)]
pub enum UserError {
    ZoneNotFound,
    NotFound,
    UserConflict,
    BadCreds,
    MissingToken,
    ExpiredSession,
    MalformedHeader,
    PermissionDenied,
    DbError(DieselError),
    PasswordError(PasswordHashError),
}

impl From<DieselError> for UserError {
    fn from(e: DieselError) -> Self {
        UserError::DbError(e)
    }
}

impl From<PasswordHashError> for UserError {
    fn from(e: PasswordHashError) -> Self {
        UserError::PasswordError(e)
    }
}


#[derive(Serialize, Debug)]
pub struct ErrorResponse {
    #[serde(with = "StatusDef")]
    #[serde(flatten)]
    pub status: Status,
    pub message: String,
    #[serde(skip_serializing_if = "Option::is_none")]
    pub details: Option<Value>
}

#[derive(Serialize)]
#[serde(remote = "Status")]
struct StatusDef {
    code: u16,
    #[serde(rename = "status", getter = "Status::reason")]
    reason: Option<&'static str>,
}

impl ErrorResponse {
    pub fn new(status: Status, message: String) -> ErrorResponse {
        ErrorResponse {
            status,
            message,
            details: None,
        }
    }

    pub fn with_details<T: Serialize> (self, details: T) -> ErrorResponse {
        ErrorResponse {
            details: serde_json::to_value(details).ok(),
            ..self
        }
    }

    pub fn err<R>(self) -> Result<R, ErrorResponse> {
        Err(self)
    }
}

impl<'r> Responder<'r, 'static> for ErrorResponse {
    fn respond_to(self, req: &'r Request<'_>) -> response::Result<'static> {
        let status = self.status;
        Response::build_from(Json(self).respond_to(req)?).status(status).ok()
    }
}

impl From<UserError> for ErrorResponse {
    fn from(e: UserError) -> Self {
        match e {
            UserError::BadCreds => ErrorResponse::new(Status::Unauthorized, "Provided credentials or token do not match any existing user".into()),
            UserError::UserConflict => ErrorResponse::new(Status::Conflict, "This user already exists".into()),
            UserError::NotFound => ErrorResponse::new(Status::NotFound, "User does not exist".into()),
            UserError::MissingToken => ErrorResponse::new(Status::Unauthorized, "Missing authorization token".into()),
            UserError::ExpiredSession => ErrorResponse::new(Status::Unauthorized, "The provided session token has expired".into()),
            UserError::MalformedHeader => ErrorResponse::new(Status::BadRequest, "Malformed authorization header".into()),
            UserError::PermissionDenied => ErrorResponse::new(Status::Forbidden, "Bearer is not authorized to access the resource".into()),
            UserError::ZoneNotFound => ErrorResponse::new(Status::NotFound, "DNS zone does not exist".into()),
            UserError::DbError(e) => make_500(e),
            UserError::PasswordError(e) => make_500(e)
        }
    }
}

impl From<Box<dyn ConnectorError>> for ErrorResponse {
    fn from(e: Box<dyn ConnectorError>) -> Self {
        if e.is_proto_error() {
            error!("{}", e);
            return make_500(e);
        } else {
            warn!("{}", e);
            let error = ErrorResponse::new(
                Status::NotFound,
                "Zone could not be found".into()
            );
            if let Some(zone) = e.zone_name() {
                return error.with_details(json!({
                    "zone_name": zone.to_utf8()
                }));
            } else {
                return error;
            }
        }
    }
}

impl From<models::RecordListParseError> for ErrorResponse {
    fn from(e: models::RecordListParseError) -> Self {
        match e {
            models::RecordListParseError::RecordNotInZone { zone, class, mismatched_class, mismatched_zone} => {
                ErrorResponse::new(
                    Status::BadRequest,
                    "Record list contains records that do not belong to the zone".into()
                ).with_details(
                    json!({
                        "zone_name": zone.to_utf8(),
                        "class": models::DNSClass::from(class),
                        "mismatched_class": mismatched_class,
                        "mismatched_zone": mismatched_zone,
                    })
                )
            },
            models::RecordListParseError::ParseError { zone, bad_records } => {
                ErrorResponse::new(
                    Status::BadRequest,
                    "Record list contains records that could not be parsed into DNS records".into()
                ).with_details(
                    json!({
                        "zone_name": zone.to_utf8(),
                        "records": bad_records
                    })
                )
            }
        }
    }
}


impl<S> From<ErrorResponse> for Outcome<S, ErrorResponse> {
    fn from(e: ErrorResponse) -> Self {
        Outcome::Failure(e.into())
    }
}


impl From<ErrorResponse> for (Status, ErrorResponse) {
    fn from(e: ErrorResponse) -> Self {
        (e.status, e)
    }
}

// TODO: change for Display trait
pub fn make_500<E: std::fmt::Debug>(e: E) -> ErrorResponse {
    error!("Making 500 for Error: {:?}", e);

    ErrorResponse::new(Status::InternalServerError, "An unexpected error occured".into())
}
