use std::convert::TryFrom;
use std::net::{Ipv6Addr, Ipv4Addr};

use base64::{Engine, engine::general_purpose};
use serde::{Deserialize, Serialize};

use trust_dns_client::serialize::binary::BinEncoder;
use trust_dns_proto::error::ProtoError;

use crate::dns;
use super::name::SerdeName;


#[derive(Debug, Deserialize, Serialize, Clone)]
#[serde(tag = "Type")]
#[serde(rename_all = "UPPERCASE")]
pub enum RData {
    #[serde(rename_all = "PascalCase")]
    A {
        address: Ipv4Addr
    },
    #[serde(rename_all = "PascalCase")]
    AAAA {
        address: Ipv6Addr
    },
    #[serde(rename_all = "PascalCase")]
    CAA {
        issuer_critical: bool,
        value: String,
        property_tag: String,
    },
    #[serde(rename_all = "PascalCase")]
    CNAME {
        target: SerdeName
    },
    // HINFO(HINFO),
    // HTTPS(SVCB),
    #[serde(rename_all = "PascalCase")]
    MX {
        preference: u16,
        mail_exchanger: SerdeName
    },
    // NAPTR(NAPTR),
    #[serde(rename_all = "PascalCase")]
    NULL {
        data: String
    },
    #[serde(rename_all = "PascalCase")]
    NS {
        target: SerdeName
    },
    // OPENPGPKEY(OPENPGPKEY),
    // OPT(OPT),
    #[serde(rename_all = "PascalCase")]
    PTR {
        target: SerdeName
    },
    #[serde(rename_all = "PascalCase")]
    SOA {
        master_server_name: SerdeName,
        maintainer_name: SerdeName,
        refresh: i32,
        retry: i32,
        expire: i32,
        minimum: u32,
        serial: u32
    },
    #[serde(rename_all = "PascalCase")]
    SRV {
        server: SerdeName,
        port: u16,
        priority: u16,
        weight: u16,
    },
    #[serde(rename_all = "PascalCase")]
    SSHFP {
        algorithm: u8,
        digest_type: u8,
        fingerprint: String,
    },
    // SVCB(SVCB),
    // TLSA(TLSA),
    #[serde(rename_all = "PascalCase")]
    TXT {
        text: String
    },

    // TODO: Eventually allow deserialization of DNSSEC records
    #[serde(skip)]
    DNSSEC(dns::DNSSECRData),
    #[serde(rename_all = "PascalCase")]
    Unknown {
        code: u16,
        data: String,
    },
    // ZERO,

    // TODO: DS (added in https://github.com/bluejekyll/trust-dns/pull/1635)
    // TODO: TLSA
}

impl From<dns::RData> for RData {
    fn from(rdata: dns::RData) -> RData {
        match rdata {
            dns::RData::A(address) => RData::A { address },
            dns::RData::AAAA(address) => RData::AAAA { address },
            // Still a draft, no iana number yet, I don't to put something that is not currently supported so that's why NULL and not unknown.
            // TODO: probably need better error here, I don't know what to do about that as this would require to change the From for something else.
            // (empty data because I'm lazy)
            dns::RData::ANAME(_) =>  RData::NULL {
                data: String::new()
            },
            dns::RData::CNAME(target) => RData::CNAME {
                target: SerdeName(target)
            },
            dns::RData::CAA(caa) => {
                let value_str = caa.value().to_string();

                RData::CAA {
                    issuer_critical: caa.issuer_critical(),
                    // Remove first and last char (byte) because string is quoted (") (should be a safe operation)
                    value: value_str[1..(value_str.len())].into(),
                    property_tag: caa.tag().as_str().to_string(),
                }
            },
            dns::RData::MX(mx) => RData::MX {
                preference: mx.preference(),
                mail_exchanger: SerdeName(mx.exchange().clone())
            },
            dns::RData::NULL(null) =>  RData::NULL {
                data: general_purpose::STANDARD.encode(null.anything())
            },
            dns::RData::NS(target) => RData::NS {
                target: SerdeName(target)
            },
            dns::RData::PTR(target) => RData::PTR {
                target: SerdeName(target)
            },
            dns::RData::SOA(soa) => RData::SOA {
                master_server_name: SerdeName(soa.mname().clone()),
                maintainer_name: SerdeName(soa.rname().clone()),
                refresh: soa.refresh(),
                retry: soa.retry(),
                expire: soa.expire(),
                minimum: soa.minimum(),
                serial: soa.serial()
            },
            dns::RData::SRV(srv) => RData::SRV {
                server: SerdeName(srv.target().clone()),
                port: srv.port(),
                priority: srv.priority(),
                weight: srv.weight(),
            },
            dns::RData::SSHFP(sshfp) => RData::SSHFP {
                algorithm: sshfp.algorithm().into(),
                digest_type: sshfp.fingerprint_type().into(),
                fingerprint: dns::sshfp::HEX.encode(sshfp.fingerprint()),
            },
            //TODO: This might alter data if not utf8 compatible, probably need to be replaced
            //TODO: check whether concatenating txt data is harmful or not
            dns::RData::TXT(txt) => RData::TXT { text: format!("{}", txt) },
            dns::RData::DNSSEC(data) => RData::DNSSEC(data),
            rdata => {
                let code = rdata.to_record_type().into();
                let mut data = Vec::new();
                let mut encoder = BinEncoder::new(&mut data);
                // TODO: need better error handling (use TryFrom ?)
                rdata.emit(&mut encoder).expect("could not encode data");

                RData::Unknown {
                    code,
                    data: general_purpose::STANDARD.encode(data),
                }
            }
        }
    }
}

impl TryFrom<RData> for dns::RData {
    type Error = ProtoError;

    fn try_from(rdata: RData) -> Result<Self, Self::Error> {
        Ok(match rdata {
            RData::A { address } => dns::RData::A(address),
            RData::AAAA { address } => dns::RData::AAAA(address),
            // TODO: Round trip test all types below (currently not tested...)
            RData::CAA { issuer_critical, value, property_tag } => {
                let property = dns::caa::Property::from(property_tag);
                let caa_value = {
                    // TODO: duplicate of trust_dns_client::serialize::txt::rdata_parser::caa::parse
                    // because caa::read_value is private
                    match property {
                        dns::caa::Property::Issue | dns::caa::Property::IssueWild => {
                            let value = dns::caa::read_issuer(value.as_bytes())?;
                            dns::caa::Value::Issuer(value.0, value.1)
                        }
                        dns::caa::Property::Iodef => {
                            let url = dns::caa::read_iodef(value.as_bytes())?;
                            dns::caa::Value::Url(url)
                        }
                        dns::caa::Property::Unknown(_) => dns::caa::Value::Unknown(value.as_bytes().to_vec()),
                    }
                };
                dns::RData::CAA(dns::caa::CAA {
                    issuer_critical,
                    tag: property,
                    value: caa_value,
                })
            },
            RData::CNAME { target } => dns::RData::CNAME(target.into_inner()),
            RData::MX { preference, mail_exchanger } => dns::RData::MX(
                dns::mx::MX::new(preference, mail_exchanger.into_inner())
            ),
            RData::NULL { data } => dns::RData::NULL(
                dns::null::NULL::with(
                    general_purpose::STANDARD.decode(data).map_err(|e| ProtoError::from(format!("{}", e)))?
                )
            ),
            RData::NS { target } => dns::RData::NS(target.into_inner()),
            RData::PTR { target } => dns::RData::PTR(target.into_inner()),
            RData::SOA {
                master_server_name,
                maintainer_name,
                refresh,
                retry,
                expire,
                minimum,
                serial
            } => dns::RData::SOA(
                dns::soa::SOA::new(
                    master_server_name.into_inner(),
                    maintainer_name.into_inner(),
                    serial,
                    refresh,
                    retry,
                    expire,
                    minimum,
                )
            ),
            RData::SRV { server, port, priority, weight } => dns::RData::SRV(
                dns::srv::SRV::new(priority, weight, port, server.into_inner())
            ),
            RData::SSHFP { algorithm, digest_type, fingerprint } => dns::RData::SSHFP(
                dns::sshfp::SSHFP::new(
                    // NOTE: This allows unassigned algorithms
                    dns::sshfp::Algorithm::from(algorithm),
                    dns::sshfp::FingerprintType::from(digest_type),
                    dns::sshfp::HEX.decode(fingerprint.as_bytes()).map_err(|e| ProtoError::from(format!("{}", e)))?
                )
            ),
            RData::TXT { text } => dns::RData::TXT(dns::txt::TXT::new(vec![text])),
            // TODO: Error out for DNSSEC? Prefer downstream checks?
            RData::DNSSEC(_) => todo!(),
            // TODO: Disallow unknown? (could be used to bypass unsopported types?) Prefer downstream checks?
            RData::Unknown { code: _code, data: _data } => todo!(),
        })
    }
}
