use std::convert::{TryFrom, TryInto};
use serde::{Deserialize, Serialize};
use trust_dns_proto::error::ProtoError;

use crate::dns;
use super::name::SerdeName;
use super::class::DNSClass;
use super::rdata::RData;


#[derive(Debug, Deserialize, Serialize, Clone)]
pub struct Record {
    #[serde(rename = "Name")]
    pub name: SerdeName,
    // TODO: Make class optional, default to IN
    #[serde(rename = "Class")]
    pub dns_class: DNSClass,
    #[serde(rename = "TTL")]
    pub ttl: u32,
    #[serde(flatten)]
    pub rdata: RData,
}

impl From<dns::Record> for Record {
    fn from(record: dns::Record) -> Record {
        Record {
            name: SerdeName(record.name().clone()),
            dns_class: record.dns_class().into(),
            ttl: record.ttl(),
            // Assume data exists, record with empty data should be filtered by caller
            rdata: record.into_data().unwrap().into(),
        }
    }
}

impl TryFrom<Record> for dns::Record {
    type Error = ProtoError;

    fn try_from(record: Record) -> Result<Self, Self::Error> {
        let mut trust_dns_record = dns::Record::from_rdata(record.name.into_inner(), record.ttl, record.rdata.try_into()?);
        trust_dns_record.set_dns_class(record.dns_class.into());
        Ok(trust_dns_record)
    }
}


pub type RecordList = Vec<Record>;

#[derive(Debug, Deserialize)]
#[serde(rename_all = "camelCase")]
pub struct UpdateRecordsRequest {
    pub old_records: RecordList,
    pub new_records: RecordList,
}

pub enum RecordListParseError {
    ParseError {
        bad_records: Vec<Record>,
        zone: dns::Name,
    },
    RecordNotInZone {
        zone: dns::Name,
        class: dns::DNSClass,
        mismatched_class: Vec<Record>,
        mismatched_zone: Vec<Record>,
    },
}

pub trait ParseRecordList {
    fn try_into_dns_type(self, zone: dns::Name, class: dns::DNSClass) -> Result<Vec<dns::Record>, RecordListParseError>;
}

impl ParseRecordList for RecordList {
    fn try_into_dns_type(self, zone: dns::Name, class: dns::DNSClass) -> Result<Vec<dns::Record>, RecordListParseError> {
        // TODO: What about relative names (also in cnames and stuff)
        let mut bad_records = Vec::new();
        let mut records: Vec<dns::Record> = Vec::new();
        let mut mismatched_class: Vec<Record> = Vec::new();
        let mut mismatched_zone: Vec<Record> = Vec::new();

        for record in self.into_iter() {
            let this_record = record.clone();
            if let Ok(record) = dns::Record::try_from(record) {
                let mut good_record = true;

                if !zone.zone_of(record.name()) {
                    mismatched_zone.push(this_record.clone());
                    good_record = false;
                }

                if record.dns_class() != class {
                    mismatched_class.push(this_record.clone());
                    good_record = false;
                }

                if good_record {
                    records.push(record);
                }
            } else {
                bad_records.push(this_record.clone());
            }
        }

        if !bad_records.is_empty() {
            return Err(RecordListParseError::ParseError {
                zone,
                bad_records,
            });
        }

        if !mismatched_class.is_empty() || !mismatched_zone.is_empty() {
            return Err(RecordListParseError::RecordNotInZone {
                zone,
                class,
                mismatched_zone,
                mismatched_class
            });
        }

        return Ok(records)
    }
}
